from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def selavy(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    """
    The selavy program is used for source finding and characterisation
    """
    if parset is None:
        parset = Parset(prefix="Selavy")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Selavy")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    selavy_app = parset_app(
        name="selavy",
        parset=parset,
        out_parset_name="taitale_selavy.in",
        cmd="selavy -c {parset_name}",
        mpi_compartible=True,
    )

    selavy_app(**kwargs)


selavy.__doc__ = """
The selavy program is used for source finding and characterisation

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Selavy".
**kwargs : dict
    Additional keyword arguments passed to the selavy application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `selavy <https://yandasoft.readthedocs.io/en/latest/analysis/selavy.html>`_

Example
-------

>>> from taitale.askap import selavy
>>> # Using a parset file
>>> selavy(
...    parset="selavy.in",
...     args={
...         "image": "image.fits",
...         "snrCut": "5",
...         "flagGrowth": "true",
...         "growthThreshold": "3",
...         "flagNegative": "true",
...         "thresholdType": "sigma",
...         "verbose": "true",
...         "nsubx": "6",
...         "nsuby": "3",
...         "overlapx": "10",
...         "overlapy": "10",
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Selavy")
>>> parset.set("image", "image.fits")
>>> parset.set("snrCut", "5")
>>> parset.set("flagGrowth", "true")
>>> selavy(parset=parset)
"""



from .cbpcalibrator import cbpcalibrator
from .ccalapply import ccalapply
from .ccalibrator import ccalibrator
from .ccontsubtract import ccontsubtract
from .cddcalibrator import cddcalibrator
from .cmodel import cmodel
from .csimulator import csimulator
from .imager import imager
from .imcontsub import imcontsub
from .linmos import linmos
from .msconcat import msconcat
from .mslist import mslist
from .msmerge import msmerge
from .mssplit import mssplit
from .selavy import selavy

__all__ = [
    "imager",
    "csimulator",
    "ccalibrator",
    "ccontsubtract",
    "selavy",
    "cmodel",
    "ccalapply",
    "cbpcalibrator",
    "cddcalibrator",
    "imcontsub",
    "linmos",
    "msconcat",
    "mslist",
    "msmerge",
    "mssplit",
]



from dataclasses import dataclass, field


@dataclass
class ArgumentSet:
    args: dict = field(default_factory=dict)

    def get(self, key):
        return self.args[key]

    def set(self, key: str, value: str):
        self.args[key] = value

    def update(self, **kwargs):
        self.args.update(kwargs)

    def merge(self, other: "ArgumentSet"):
        self.args = {**self.args, **other.args}



from taitale.utils import CliArgs
from taitale.utils.wrapper import cli_app


def wsclean(**kwargs):
    cli_args = CliArgs()
    wsclean_app = cli_app(
        name="wsclean",
        cmd="wsclean {option}",
        cli_args=cli_args,
    )

    wsclean_app(**kwargs)


# TODO: Add doc strings

wsclean.__doc__ = """
Taitale interface to wsclean.

Notes
-----

For further information on usage check the WSClean documentation for `wsclean <https://wsclean.readthedocs.io/en/latest/getting_started.html>`_

Example
-------

>>> from taitale.lofar import wsclean
>>> wsclean(
...     args="wsclean -no-update-model-required -verbose -reorder \\
...         -size 4096 4096 -scale 2arcsec -pol QU -mgain 0.85 -niter 50000 \\
...         -auto-threshold 3 -join-polarizations -squared-channel-joining -log-time \\
...         -no-mf-weighting -name output input.ms"
... )
"""



import argparse

from taitale.utils import CliArgs
from taitale.utils.wrapper import cli_app


def msmerge_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output-file")
    parser.add_argument("-x", "--tileNcorr")
    parser.add_argument("-c", "--tileNchan")
    parser.add_argument("-r", "--tileNrow")
    parser.add_argument("-b", "--blocksize")
    parser.add_argument("-B", "--bufferMB")
    parser.add_argument("-f", "--format")
    parser.add_argument("-odirect", action="store_true")
    parser.add_argument("-d", "--dryrun", action="store_true")
    parser.add_argument("--full", action="store_true")
    parser.add_argument("-p", "--parameter-file")
    parser.add_argument("-i", "--input-file")

    parser.add_argument("positional", nargs="+")
    return parser


def msmerge(**kwargs):
    parser = msmerge_arguments()
    cli_args = CliArgs(parser=parser)
    msmerge_app = cli_app(
        name="msmerge",
        cmd="msmerge {option}",
        cli_args=cli_args,
    )

    msmerge_app(**kwargs)


msmerge.__doc__ = """
The msmerge program is used to merge multiple measurement sets.

Parameters
----------
**kwargs : dict
    Additional keyword arguments passed to the msmerge application.

Notes
-----
For further information on usage, check the ASKAPsoft documentation for `msmerge <https://yandasoft.readthedocs.io/en/latest/utils/msmerge.html>`_

Example
-------
>>> from taitale.askap import msmerge
>>> msmerge(
...     args={
...         "output_file": "merged.ms",
...         "tileNcorr": "4",
...         "tileNchan": "54",
...         "positional": ["input1.ms", "input2.ms", "input3.ms"]
...     }
... )
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def cmodel(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Cmodel")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Cmodel")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    cmodel_app = parset_app(
        name="cmodel",
        parset=parset,
        out_parset_name="taitale_cmodel.in",
        cmd="cmodel -c {parset_name}",
        mpi_compartible=True,
    )

    cmodel_app(**kwargs)


cmodel.__doc__ = """
The cmodel program is used to generate model visibilities

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Cmodel".
**kwargs : dict
    Additional keyword arguments passed to the cmodel application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `cmodel <https://yandasoft.readthedocs.io/en/latest/calim/cmodel.html>`_

Example
-------
>>> from taitale.askap import cmodel
>>> # Using a parset file
>>> cmodel(
...    parset="cmodel.in",
...     workers=2,
...     args={
...         "dataset": "./1934-638_0.ms",
...         "modelimage": "1934-638.model",
...         "nchan": "16416",
...         "singleoutputms": "true",
...         "useweightsderived": "true",
...         "visweightcutoff": "1e-4"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Cmodel")
>>> parset.set("dataset", "./1934-638_0.ms")
>>> cmodel(parset=parset, workers=2)
"""



import datetime
import subprocess
import time
from shlex import quote

from taitale.utils.argumentset import ArgumentSet


def getstamp():
    return datetime.datetime.fromtimestamp(time.time()).strftime("%Y-%m-%d-%H-%M-%S")


def executor(appname, run_cmd, config, mpi_compartible, logfile=None, **kwargs):
    if mpi_compartible:
        workers = kwargs.get("workers")
        if workers is not None:
            workload_manager = config["workload_manager"]
            if workload_manager != "mpi" and workload_manager != "slurm":
                raise ValueError("Given workload distributor is not supported")

            # TODO: A lot of was this can go wrong, make it simple and fail proof
            nodes = kwargs.get("nodes", 1)
            run_cmd = config[workload_manager]["workload_exec"].format(
                cmd=run_cmd,
                **config,
                **config[workload_manager],
                workers=quote(str(workers)),
                nodes=nodes,
            )
        else:
            raise ValueError("workers must be defined")

    timestamp = getstamp()
    if logfile is None:
        logfile = f"{appname}-{timestamp}.log"

    if config["runtime"] == "container":
        container_run = config["container"]["exec"]
        run_cmd = container_run.format(
            workload_exec=run_cmd, logfile=quote(logfile), **config
        )
    elif config["runtime"] == "native":
        native_run = config["native"]["exec"]
        run_cmd = native_run.format(
            workload_exec=run_cmd, logfile=quote(logfile), **config
        )

    print(f"Starting {appname}")
    if config["dryrun"] is True:
        print(run_cmd)
    else:
        if config["ipy"] is True:
            from taitale.widgets import ICmdRunWithLogTrail

            status_code = ICmdRunWithLogTrail(run_cmd, logfile)
            if status_code:
                raise subprocess.CalledProcessError(status_code, run_cmd)
        else:
            out = subprocess.run(f"{run_cmd}", shell=True)
            out.check_returncode()

    print(f"{appname} complete")


def parset_app(
    name,
    parset,
    out_parset_name,
    cmd,
    # TODO: Define this flag to remove mpiexec if the user want to
    mpi_compartible=False,
    pipestderr=False,
):

    def parset_cmd_execute(env, **kwargs):
        # Filter out workers from parset update
        parset_kwargs = kwargs["args"] if "args" in kwargs else {}
        output = parset.merge(ArgumentSet(args=parset_kwargs))
        env["pipestderr"] = " 2>&1" if pipestderr else ""

        if env["dryrun"] is True:
            print(parset.to_string())
        else:
            parset.serialize(out_parset_name)

        run_cmd = cmd.format(parset_name=quote(out_parset_name))
        executor(name, run_cmd, env, mpi_compartible=mpi_compartible, **kwargs)
        return output

    return parset_cmd_execute


def cli_app(
    name,
    cli_args,
    cmd,
    mpi_compartible=False,
    pipestderr=False,
):
    def cli_cmd_execute(env, **kwargs):
        parset_kwargs = kwargs["args"] if "args" in kwargs else ""
        output = cli_args.parse_string(parset_kwargs)
        env["pipestderr"] = " 2>&1" if pipestderr else ""

        run_cmd = cmd.format(option=cli_args.to_string())
        executor(name, run_cmd, env, mpi_compartible=mpi_compartible, **kwargs)
        return output

    return cli_cmd_execute
